import sys
import urllib.request
import json
import datetime
# minimal python script that returns a strategy matrix, for given github.event_name


with urllib.request.urlopen('https://endoflife.date/api/python.json') as f:
    python_versions = json.loads(f.read().decode('utf-8'))

now = datetime.datetime.now().strftime('%Y-%m-%d')
python_supported_versions = [ v['cycle'] for v in python_versions if v['eol'] > now and v['cycle'] not in ['3.7','3.8','3.12'] ]

python_default_version = python_supported_versions[1]

matrix = {
    'default': {
        'os': ['ubuntu-latest'],
        'python_version': [python_default_version],
    },
    'pull_request': {
        'os': ['ubuntu-latest'],
        'python_version': [python_default_version],
    },
    'schedule': {
        'os': ['ubuntu-latest', 'macos-latest', 'windows-latest'],
        'python_version': python_supported_versions,
    },
    'release': {
        'os': ['ubuntu-latest', 'macos-latest', 'windows-latest'],
        'python_version': python_supported_versions,
    },
    'pull_request_review': {
        'os': ['ubuntu-latest', 'macos-latest', 'windows-latest'],
        'python_version': python_supported_versions,
    },
    'workflow_dispatch': {
        'os': ['ubuntu-latest', 'macos-latest', 'windows-latest'],
        'python_version': python_supported_versions,
    },
}

if __name__ == "__main__":
    try:
        event = sys.argv[1]
    except IndexError:
        event = 'default'
    if event not in matrix:
        event = 'default'

    print('::set-output name=os_matrix::%s' % str(matrix[event]['os']))
    print('::set-output name=python_version_matrix::%s' % str(matrix[event]['python_version']))



import xsar
import rasterio
import os
import logging
import dill
import pickle


logging.basicConfig()
logging.getLogger('xsar').setLevel(logging.DEBUG)
logging.getLogger('xsar.utils').setLevel(logging.DEBUG)
logging.getLogger('xsar.xml_parser').setLevel(logging.DEBUG)
logging.captureWarnings(True)

logger = logging.getLogger('xsar_test')
logger.setLevel(logging.DEBUG)

# logger.info('using %s as test_dir' % xsar.config['data_dir'])


meta = xsar.Sentinel1Meta(
    xsar.get_test_file('S1A_IW_GRDH_1SDV_20170907T103020_20170907T103045_018268_01EB76_Z010.SAFE'))


def test_open_dataset():
    try:
        ds = xsar.open_dataset(meta, resolution={'sample': 100, 'line': 100}, resampling=rasterio.enums.Resampling.average)
        ds.compute()
        ds = xsar.open_dataset(meta).isel(pol=0, sample=slice(0, 100), line=slice(0, 100))
        ds.compute()
        assert True
    except:
        assert False


def test_serializable_s1_meta():
    s1meta = dill.loads(dill.dumps(meta))
    assert isinstance(s1meta.coords2ll(100, 100), tuple)



import xarray as xr
import datetime
import numpy as np
import glob
from .utils import bind, url_get
import pandas as pd


def resource_strftime(resource, **kwargs):
    """
    From a resource string like '%Y/%j/myfile_%Y%m%d%H%M.nc' and a date like 'Timestamp('2018-10-13 06:23:22.317102')',
    returns a tuple composed of the closer available date and string like '/2018/286/myfile_201810130600.nc'

    If ressource string is an url (ie 'ftp://ecmwf/%Y/%j/myfile_%Y%m%d%H%M.nc'), fsspec will be used to retreive the file locally.

    Parameters
    ----------
    resource: str

        resource string, with strftime template

    date: datetime

        date to be used

    step: int

        hour step between 2 files


    Returns
    -------
    tuple : (datetime,str)

    """

    date = kwargs['date']
    step = kwargs['step']

    delta = datetime.timedelta(hours=step) / 2
    date = date.replace(
        year=(date+delta).year,
        month=(date+delta).month,
        day=(date+delta).day,
        hour=(date+delta).hour // step * step,
        minute=0,
        second=0,
        microsecond=0
    )

    return date, url_get(date.strftime(resource))


def _to_lon180(ds):
    # roll [0, 360] to [-180, 180] on dim x
    ds = ds.roll(x=-np.searchsorted(ds.x, 180), roll_coords=True)
    ds['x'] = xr.where(ds['x'] >= 180, ds['x'] - 360, ds['x'])
    return ds


def ecmwf_0100_1h(fname, **kwargs):
    """
    ecmwf 0.1 deg 1h reader (ECMWF_FORECAST_0100_202109091300_10U_10V.nc)

    Parameters
    ----------
    fname: str

        hwrf filename

    Returns
    -------
    xarray.Dataset
    """
    ecmwf_ds = xr.open_dataset(
        fname, chunks={'Longitude': 1000, 'Latitude': 1000}).isel(time=0)
    ecmwf_ds.attrs['time'] = datetime.datetime.fromtimestamp(
        ecmwf_ds.time.item() // 1000000000)
    if 'time' in ecmwf_ds:
        ecmwf_ds = ecmwf_ds.drop("time")
    ecmwf_ds = ecmwf_ds[["Longitude", "Latitude", "10U", "10V"]].rename(
        {
            'Longitude': 'x',
            'Latitude': 'y',
            '10U': 'U10',
            '10V': 'V10'
        }
    )
    ecmwf_ds.attrs = {k: ecmwf_ds.attrs[k]
                      for k in ['title', 'institution', 'time']}

    # dataset is lon [0, 360], make it [-180,180]
    ecmwf_ds = _to_lon180(ecmwf_ds)

    ecmwf_ds.rio.write_crs("EPSG:4326", inplace=True)

    return ecmwf_ds


def ecmwf_0125_1h(fname, **kwargs):
    """
    ecmwf 0.125 deg 1h reader (ecmwf_201709071100.nc)

    Parameters
    ----------
    fname: str

        hwrf filename

    Returns
    -------
    xarray.Dataset
    """
    ecmwf_ds = xr.open_dataset(
        fname, chunks={'longitude': 1000, 'latitude': 1000})

    ecmwf_ds = ecmwf_ds.rename(
        {'longitude': 'x', 'latitude': 'y'}
    ).rename(
        {'Longitude': 'x', 'Latitude': 'y', 'U': 'U10', 'V': 'V10'}
    ).set_coords(['x', 'y'])

    ecmwf_ds['x'] = ecmwf_ds.x.compute()
    ecmwf_ds['y'] = ecmwf_ds.y.compute()

    # dataset is lon [0, 360], make it [-180,180]
    ecmwf_ds = _to_lon180(ecmwf_ds)

    ecmwf_ds.attrs['time'] = datetime.datetime.fromisoformat(
        ecmwf_ds.attrs['date'])

    ecmwf_ds.rio.write_crs("EPSG:4326", inplace=True)

    return ecmwf_ds


def hwrf_0015_3h(fname, **kwargs):
    """
    hwrf 0.015 deg 3h reader ()


    Parameters
    ----------
    fname: str

        hwrf filename

    Returns
    -------
    xarray.Dataset
    """
    hwrf_ds = xr.open_dataset(fname)
    try:
        hwrf_ds = hwrf_ds[['U', 'V', 'LON', 'LAT']]
        hwrf_ds = hwrf_ds.squeeze('t', drop=True)

    except Exception as e:
        raise ValueError("date '%s' can't be find in %s " %
                         (kwargs['date'], fname))

    hwrf_ds.attrs['time'] = datetime.datetime.strftime(
        kwargs['date'], '%Y-%m-%d %H:%M:%S')

    hwrf_ds = hwrf_ds.assign_coords({"x": hwrf_ds.LON.values[0, :], "y": hwrf_ds.LAT.values[:, 0]}).drop_vars(['LON', 'LAT']).rename(
        {
            'U': 'U10',
            'V': 'V10'
        }
    )

    # hwrf_ds.attrs = {k: hwrf_ds.attrs[k] for k in ['institution', 'time']}
    hwrf_ds = _to_lon180(hwrf_ds)
    hwrf_ds.rio.write_crs("EPSG:4326", inplace=True)

    return hwrf_ds


def era5_0250_1h(fname, **kwargs):
    """
    era5 0.250 deg 1h reader ()


    Parameters
    ----------
    fname: str

        era5 filename

    Returns
    -------
    xarray.Dataset
    """

    ds_era5 = xr.open_dataset(fname, chunks={'time': 1})
    ds_era5 = ds_era5[['u10', 'v10', 'latitude025', 'longitude025']]
    ds_era5 = ds_era5.sel(time=str(kwargs['date']))
    ds_era5 = ds_era5.drop('time')

    ds_era5 = ds_era5.rename(
        {
            'longitude025': 'x',
            'latitude025': 'y',
            'u10': 'U10',
            'v10': 'V10'
        }
    )

    ds_era5.attrs['time'] = kwargs['date']
    ds_era5 = _to_lon180(ds_era5)
    ds_era5.rio.write_crs("EPSG:4326", inplace=True)
    return ds_era5


def gebco(gebco_files):
    """gebco file reader (geotiff from https://www.gebco.net/data_and_products/gridded_bathymetry_data)"""
    return xr.combine_by_coords(
        [
            xr.open_dataset(
                f, chunks={'x': 1000, 'y': 1000}
            ).isel(band=0).drop_vars('band') for f in gebco_files
        ]
    )


# list available rasters as a pandas dataframe
available_rasters = pd.DataFrame(
    columns=['resource', 'read_function', 'get_function'])
available_rasters.loc['gebco'] = [None, gebco, glob.glob]
available_rasters.loc['ecmwf_0100_1h'] = [
    None, ecmwf_0100_1h, bind(resource_strftime, ..., step=1)]
available_rasters.loc['ecmwf_0125_1h'] = [
    None, ecmwf_0125_1h, bind(resource_strftime, ..., step=1)]
available_rasters.loc['hwrf_0015_3h'] = [
    None, hwrf_0015_3h, bind(resource_strftime, ..., step=3)]
available_rasters.loc['era5_0250_1h'] = [
    None, era5_0250_1h, bind(resource_strftime, ..., step=1)]



# -*- coding: utf-8 -*-
import logging
import warnings


from .radarsat2_meta import RadarSat2Meta
from .utils import timing, map_blocks_coords, BlockingActorProxy, to_lon180
import numpy as np
import rasterio.features
import xarray as xr
from scipy.interpolate import RectBivariateSpline, interp1d
import dask
import datatree
from .base_dataset import BaseDataset

logger = logging.getLogger('xsar.radarsat2_dataset')
logger.addHandler(logging.NullHandler())

# we know tiff as no geotransform : ignore warning
warnings.filterwarnings(
    "ignore", category=rasterio.errors.NotGeoreferencedWarning)

# allow nan without warnings
# some dask warnings are still non filtered: https://github.com/dask/dask/issues/3245
np.errstate(invalid='ignore')


class RadarSat2Dataset(BaseDataset):
    """
        Handle a SAFE subdataset.
        A dataset might contain several tiff files (multiples polarizations), but all tiff files must share the same footprint.

        The main attribute useful to the end-user is `self.dataset` (`xarray.Dataset` , with all variables parsed from xml and tiff files.)

        Parameters
        ----------
        dataset_id: str or RadarSat2Meta object

            if str, it can be a path, or a gdal dataset identifier)

        resolution: dict, number or string, optional
            resampling dict like `{'line': 20, 'sample': 20}` where 20 is in pixels.

            if a number, dict will be constructed from `{'line': number, 'sample': number}`

            if str, it must end with 'm' (meters), like '100m'. dict will be computed from sensor pixel size.

        resampling: rasterio.enums.Resampling or str, optional

            Only used if `resolution` is not None.

            ` rasterio.enums.Resampling.rms` by default. `rasterio.enums.Resampling.nearest` (decimation) is fastest.

        chunks: dict, optional

            dict with keys ['pol','line','sample'] (dask chunks).

        dtypes: None or dict, optional

            Specify the data type for each variable.

        lazyloading: bool, optional
            activate or not the lazy loading of the high resolution fields

    """

    def __init__(self, dataset_id, resolution=None,
                 resampling=rasterio.enums.Resampling.rms,
                 chunks={'line': 5000, 'sample': 5000},
                 dtypes=None, lazyloading=True, skip_variables=None):
        if skip_variables is None:
            skip_variables = []
        # default dtypes
        if dtypes is not None:
            self._dtypes.update(dtypes)

        # default meta for map_blocks output.
        # as asarray is imported from numpy, it's a numpy array.
        # but if later we decide to import asarray from cupy, il will be a cupy.array (gpu)
        self.sar_meta = None
        self.resolution = resolution

        if not isinstance(dataset_id, RadarSat2Meta):
            self.sar_meta = BlockingActorProxy(RadarSat2Meta, dataset_id)
            # check serializable
            # import pickle
            # s1meta = pickle.loads(pickle.dumps(self.s1meta))
            # assert isinstance(sar_meta.coords2ll(100, 100),tuple)
        else:
            # we want self.sar_meta to be a dask actor on a worker
            self.sar_meta = BlockingActorProxy(
                RadarSat2Meta.from_dict, dataset_id.dict)
        del dataset_id
        if self.sar_meta.multidataset:
            raise IndexError(
                """Can't open an multi-dataset. Use `xsar.RadarSat2Meta('%s').subdatasets` to show availables ones""" % self.sar_meta.path
            )

        from xradarsat2 import load_digital_number
        # build datatree
        DN_tmp = load_digital_number(self.sar_meta.dt, resolution=resolution,
                                     resampling=resampling, chunks=chunks)['digital_numbers'].ds
        # In order to respect xsar convention, lines and samples have been flipped in the metadata when necessary.
        # `load_digital_number` uses these metadata but rio creates new coords without keeping the flipping done.
        # So we have to flip again a time digital numbers to respect xsar convention
        DN_tmp = self.flip_sample_da(DN_tmp)
        DN_tmp = self.flip_line_da(DN_tmp)

        # geoloc
        geoloc = self.sar_meta.geoloc
        geoloc.attrs['history'] = 'annotations'

        # orbitAndAttitude
        orbit_and_attitude = self.sar_meta.orbit_and_attitude
        orbit_and_attitude.attrs['history'] = 'annotations'

        # dopplerCentroid
        doppler_centroid = self.sar_meta.doppler_centroid
        doppler_centroid.attrs['history'] = 'annotations'

        # dopplerRateValues
        doppler_rate_values = self.sar_meta.doppler_rate_values
        doppler_rate_values.attrs['history'] = 'annotations'

        # chirp
        chirp = self.sar_meta.chirp
        chirp.attrs['history'] = 'annotations'

        # radarParameters
        radar_parameters = self.sar_meta.radar_parameters
        radar_parameters.attrs['history'] = 'annotations'

        # lookUpTables
        lut = self.sar_meta.lut
        lut.attrs['history'] = 'annotations'

        self.datatree = datatree.DataTree.from_dict({'measurement': DN_tmp, 'geolocation_annotation': geoloc
                                                     })

        self._dataset = self.datatree['measurement'].to_dataset()

        # dict mapping for variable names to create by applying specified lut on digital numbers

        self._map_var_lut_noise = {
            'sigma0': 'noiseLevelValues_SigmaNought',
            'gamma0': 'noiseLevelValues_Gamma',
            'beta0': 'noiseLevelValues_BetaNought',
        }

        self._map_var_lut = {
            'sigma0': 'lutSigma',
            'gamma0': 'lutGamma',
            'beta0': 'lutBeta',
        }
        geoloc_vars = ['latitude', 'longitude', 'altitude',
                       'incidence', 'elevation']
        for vv in skip_variables:
            if vv in geoloc_vars:
                geoloc_vars.remove(vv)

        for att in ['name', 'short_name', 'product', 'safe', 'swath', 'multidataset']:
            if att not in self.datatree.attrs:
                # tmp = xr.DataArray(self.s1meta.__getattr__(att),attrs={'source':'filename decoding'})
                self.datatree.attrs[att] = self.sar_meta.__getattr__(att)
                self._dataset.attrs[att] = self.sar_meta.__getattr__(att)

        value_res_line = self.sar_meta.geoloc.line.attrs['rasterAttributes_sampledLineSpacing_value']
        value_res_sample = self.sar_meta.geoloc.pixel.attrs[
            'rasterAttributes_sampledPixelSpacing_value']
        # self._load_incidence_from_lut()
        refe_spacing = 'slant'
        if resolution is not None:
            # if the data sampling changed it means that the quantities are projected on ground
            refe_spacing = 'ground'
            if isinstance(resolution, str):
                value_res_sample = float(resolution.replace('m', ''))
                value_res_line = value_res_sample
            elif isinstance(resolution, dict):
                value_res_sample = self.sar_meta.geoloc.pixel.attrs['rasterAttributes_sampledPixelSpacing_value'] \
                    * resolution['sample']
                value_res_line = self.sar_meta.geoloc.line.attrs['rasterAttributes_sampledLineSpacing_value'] \
                    * resolution['line']
            else:
                logger.warning('resolution type not handle (%s) should be str or dict -> sampleSpacing'
                               ' and lineSpacing are not correct', type(resolution))
        self._dataset['sampleSpacing'] = xr.DataArray(value_res_sample,
                                                      attrs={'units': 'm', 'referential': refe_spacing})
        self._dataset['lineSpacing'] = xr.DataArray(
            value_res_line, attrs={'units': 'm'})

        # dataset no-pol template for function evaluation on coordinates (*no* values used)
        # what's matter here is the shape of the image, not the values.
        with warnings.catch_warnings():
            # warnings.simplefilter("ignore", np.ComplexWarning)
            self._da_tmpl = xr.DataArray(
                dask.array.empty_like(
                    self._dataset.digital_number.isel(pol=0).drop('pol'),
                    dtype=np.int8, name="empty_var_tmpl-%s" % dask.base.tokenize(self.sar_meta.name)),
                dims=('line', 'sample'),
                coords={'line': self._dataset.digital_number.line,
                        'sample': self._dataset.digital_number.sample}
            )
        # Add vars to define if lines or samples have been flipped to respect xsar convention
        self._dataset = xr.merge([
            xr.DataArray(data=self.sar_meta.samples_flipped,
                         attrs={'meaning':
                                'xsar convention : increasing incidence values along samples axis'}
                         ).to_dataset(name='samples_flipped'),
            self._dataset
        ])
        self._dataset = xr.merge([
            xr.DataArray(data=self.sar_meta.lines_flipped,
                         attrs={'meaning':
                                'xsar convention : increasing time along line axis '
                                '(whatever ascending or descending pass direction)'}
                         ).to_dataset(name='lines_flipped'),
            self._dataset
        ])

        self._luts = self.lazy_load_luts()
        self.apply_calibration_and_denoising()
        self._dataset = xr.merge([self.load_from_geoloc(
            geoloc_vars, lazy_loading=lazyloading), self._dataset])

        # compute offboresight in self._dataset
        self._get_offboresight_from_elevation()
        rasters = self._load_rasters_vars()
        if rasters is not None:
            self._dataset = xr.merge([self._dataset, rasters])
        self._dataset = xr.merge([self.interpolate_times, self._dataset])
        if 'ground_heading' not in skip_variables:
            self._dataset = xr.merge(
                [self.load_ground_heading(), self._dataset])
        if 'velocity' not in skip_variables:
            self._dataset = xr.merge(
                [self.get_sensor_velocity(), self._dataset])
        self._rasterized_masks = self.load_rasterized_masks()
        self._dataset = xr.merge([self._rasterized_masks, self._dataset])
        """a = self._dataset.copy()
        self._dataset = self.flip_sample_da(a)
        self.datatree['measurement'] = self.datatree['measurement'].assign(self._dataset)
        a = self._dataset.copy()
        self._dataset = self.flip_line_da(a)"""
        self.datatree['measurement'] = self.datatree['measurement'].assign(
            self._dataset)
        """self.datatree = datatree.DataTree.from_dict(
            {'measurement': self.datatree['measurement'],
             'geolocation_annotation': self.datatree['geolocation_annotation'],
             'reader': self.sar_meta.dt})"""
        self._reconfigure_reader_datatree()
        self._dataset.attrs.update(self.sar_meta.to_dict("all"))
        self.datatree.attrs.update(self.sar_meta.to_dict("all"))

        self.resampled = resolution is not None

    def lazy_load_luts(self):
        """
        Lazy load luts from the reader as delayed

        Returns
        -------
        xarray.Dataset
            Contains delayed dataArrays of luts
        """
        luts_ds = self.sar_meta.dt['lut'].ds.rename({'pixel': 'sample'})
        merge_list = []
        for lut_name in luts_ds:
            lut_f_delayed = dask.delayed()(luts_ds[lut_name])
            ar = dask.array.from_delayed(
                lut_f_delayed.data, (luts_ds[lut_name].data.size,), luts_ds[lut_name].dtype)
            da = xr.DataArray(data=ar, dims=['sample'], coords={'sample': luts_ds[lut_name].sample},
                              attrs=luts_ds[lut_name].attrs)
            ds_lut_f_delayed = da.to_dataset(name=lut_name)
            merge_list.append(ds_lut_f_delayed)
        return xr.combine_by_coords(merge_list)

    @timing
    def _get_offboresight_from_elevation(self):
        """
        Compute offboresight angle.

        Returns
        -------

        """
        self._dataset["offboresight"] = self._dataset.elevation - \
            (30.1833947 * self._dataset.latitude ** 0 +
             0.0082998714 * self._dataset.latitude ** 1 -
             0.00031181534 * self._dataset.latitude ** 2 -
             0.0943533e-07 * self._dataset.latitude ** 3 +
             3.0191435e-08 * self._dataset.latitude ** 4 +
             4.968415428e-12 * self._dataset.latitude ** 5 -
             9.315371305e-13 * self._dataset.latitude ** 6) + 29.45
        self._dataset["offboresight"].attrs['comment'] = 'built from elevation angle and latitude'

    @timing
    def load_from_geoloc(self, varnames, lazy_loading=True):
        """
        Interpolate (with RectBiVariateSpline) variables from `self.sar_meta.geoloc` to `self._dataset`

        Parameters
        ----------
        varnames: list of str
            subset of variables names in `self.sar_meta.geoloc`

        Returns
        -------
        xarray.Dataset
            With interpolated variables

        """
        mapping_dataset_geoloc = {'latitude': 'latitude',
                                  'longitude': 'longitude',
                                  'incidence': 'incidenceAngle',
                                  'elevation': 'elevationAngle',
                                  'altitude': 'height',
                                  }
        da_list = []

        for varname in varnames:
            varname_in_geoloc = mapping_dataset_geoloc[varname]
            if varname == 'incidence':
                da = self._load_incidence_from_lut()
                da.name = varname
                da_list.append(da)
            elif varname == 'elevation':
                da = self._load_elevation_from_lut()
                da.name = varname
                da_list.append(da)
            else:
                if varname == 'longitude':
                    z_values = self.sar_meta.geoloc[varname]
                    if self.sar_meta.cross_antemeridian:
                        logger.debug('translate longitudes between 0 and 360')
                        z_values = z_values % 360
                else:
                    z_values = self.sar_meta.geoloc[varname_in_geoloc]
                interp_func = RectBivariateSpline(
                    self.sar_meta.geoloc.line,
                    self.sar_meta.geoloc.pixel,
                    z_values,
                    kx=1, ky=1
                )
                typee = self.sar_meta.geoloc[varname_in_geoloc].dtype
                if lazy_loading:
                    da_var = map_blocks_coords(
                        self._da_tmpl.astype(typee),
                        interp_func
                    )
                else:
                    da_val = interp_func(
                        self._dataset.digital_number.line, self._dataset.digital_number.sample)
                    da_var = xr.DataArray(data=da_val, dims=['line', 'sample'],
                                          coords={'line': self._dataset.digital_number.line,
                                                  'sample': self._dataset.digital_number.sample})
                if varname == 'longitude':
                    if self.sar_meta.cross_antemeridian:
                        da_var.data = da_var.data.map_blocks(to_lon180)

                da_var.name = varname

                # copy history
                try:
                    da_var.attrs['history'] = self.sar_meta.geoloc[varname_in_geoloc].attrs['xpath']
                except KeyError:
                    pass

                da_list.append(da_var)
        ds = xr.merge(da_list)
        return ds

    @timing
    def _load_incidence_from_lut(self):
        """
        load incidence thanks to lut. In the formula to have the incidence, we understand that it is calculated
        thanks to lut. But we can ask ourselves if we consider the denoised ones or not. In this case we have chosen to
        take the not denoised luts.
        Usually look up tables depends on luts, but we already have information about look up tables, so we determine
        incidence thanks to these.
        Reference : `Radarsat2 product format definition` 7.2

        Returns
        -------
        xarray.DataArray
            DataArray of incidence (expressed in degrees)
        """
        beta = self._dataset.beta0_raw[0]
        gamma = self._dataset.gamma0_raw[0]
        incidence_pre = gamma / beta
        i_angle = np.degrees(np.arctan(incidence_pre))
        return xr.DataArray(data=i_angle, dims=['line', 'sample'], coords={
            'line': self._dataset.digital_number.line,
            'sample': self._dataset.digital_number.sample})

    @timing
    def _resample_lut_values(self, lut):
        lines = self.sar_meta.geoloc.line
        samples = np.arange(lut.shape[0])
        lut_values_2d = dask.delayed(np.tile)(lut, (lines.shape[0], 1))
        interp_func = dask.delayed(RectBivariateSpline)(
            x=lines, y=samples, z=lut_values_2d, kx=1, ky=1)
        """var = inter_func(self._dataset.digital_number.line, self._dataset.digital_number.sample)
        da_var = xr.DataArray(data=var, dims=['line', 'sample'],
                              coords={'line': self._dataset.digital_number.line,
                                      'sample': self._dataset.digital_number.sample})"""
        da_var = map_blocks_coords(
            self._da_tmpl.astype(lut.dtype),
            interp_func
        )
        return da_var

    @timing
    def _load_elevation_from_lut(self):
        """
        Load elevation from lut.
        Formula reference : `RSI-GS-026 RS-1 Data Products Specifications` 5.3.3.2.
        this formula needs the orbit altitude. But 2 variables look like this one : `satelliteHeight` and `Altitude`.
        We considered the satelliteHeight.

        Returns
        -------

        """
        satellite_height = self.sar_meta.dt.attrs['satelliteHeight']
        earth_radius = 6.371e6
        incidence = self._load_incidence_from_lut()
        angle_rad = np.sin(np.radians(incidence))
        inside = angle_rad * earth_radius / (earth_radius + satellite_height)
        return np.degrees(np.arcsin(inside))

    @timing
    def _get_lut_noise(self, var_name):
        """
        Get noise lut in the reader for var_name

        Parameters
        ----------
        var_name: str

        Returns
        -------
        xarray.DataArray
            noise lut for `var_name`
        """
        try:
            lut_name = self._map_var_lut_noise[var_name]
        except KeyError:
            raise ValueError(
                "can't find noise lut name for var '%s'" % var_name)
        try:
            lut = self.sar_meta.dt['radarParameters'][lut_name]
        except KeyError:
            raise ValueError(
                "can't find noise lut from name '%s' for variable '%s'" % (lut_name, var_name))
        return lut

    @timing
    def _interpolate_for_noise_lut(self, var_name):
        """
        Interpolate the noise level values (from the reader) and resample it to create a noise lut.
        Initial values are at low resolution, and the high resolution range is made from the pixel first noise
        level value and the step. Then, an interpolation with RectBivariateSpline permit having a full resolution
        and extrapolate the first pixels; getting by the end resampled noise values.
        Nb : Noise Level Values extracted from the reader are already calibrated, and expressed in dB
        (so they are converted in linear).

        Parameters
        ----------
        var_name : str
            Variable name to compute by applying lut. Must exist in `self._map_var_lut_noise` to be able to get the corresponding lut.

        Returns
        -------
        xarray.DataArray
            Noise level values interpolated and resampled
        """
        initial_lut = self._get_lut_noise(var_name)
        first_pix = initial_lut.attrs['pixelFirstNoiseValue']
        step = initial_lut.attrs['stepSize']
        noise_values = (10 ** (initial_lut / 10))
        lines = np.arange(self.sar_meta.geoloc.line[-1] + 1)
        noise_values_2d = np.tile(noise_values, (lines.shape[0], 1))
        indexes = [first_pix + step *
                   i for i in range(0, noise_values.shape[0])]
        interp_func = dask.delayed(RectBivariateSpline)(
            x=lines, y=indexes, z=noise_values_2d, kx=1, ky=1)
        da_var = map_blocks_coords(
            self._da_tmpl.astype(self._dtypes['noise_lut']),
            interp_func
        )
        return da_var

    @timing
    def _get_noise(self, var_name):
        """
            Get noise equivalent for  `var_name`.

            Parameters
            ----------
            var_name: str Variable name to compute. Must exist in `self._map_var_lut` and
            `self._map_var_lut_noise` to be able to get the corresponding lut.

            Returns
            -------
            xarray.Dataset
                with one variable named by `'ne%sz' % var_name[0]` (ie 'nesz' for 'sigma0', 'nebz' for 'beta0', etc...)
        """
        name = 'ne%sz' % var_name[0]
        concat_list = []
        # add pol dim
        for pol in self._dataset.pol:
            lut_noise = self._interpolate_for_noise_lut(
                var_name).assign_coords(pol=pol).expand_dims('pol')
            concat_list.append(lut_noise)
        return xr.concat(concat_list, dim='pol').to_dataset(name=name)

    def apply_calibration_and_denoising(self):
        """
        apply calibration and denoising functions to get high resolution sigma0 , beta0 and gamma0 + variables *_raw

        Returns:
        --------

        """
        for var_name, lut_name in self._map_var_lut.items():
            if lut_name in self._luts:
                # merge var_name into dataset (not denoised)
                self._dataset = self._dataset.merge(
                    self._apply_calibration_lut(var_name))
                # merge noise equivalent for var_name (named 'ne%sz' % var_name[0)
                self._dataset = self._dataset.merge(self._get_noise(var_name))
            else:
                logger.debug("Skipping variable '%s' ('%s' lut is missing)" % (
                    var_name, lut_name))
        self._dataset = self._add_denoised(self._dataset)
        self.datatree['measurement'] = self.datatree['measurement'].assign(
            self._dataset)
        # self._dataset = self.datatree[
        #     'measurement'].to_dataset()  # test oct 22 to see if then I can modify variables of the dt
        return

    def _add_denoised(self, ds, clip=False, vars=None):
        """add denoised vars to dataset

        Parameters
        ----------
        ds : xarray.DataSet
            dataset with non denoised vars, named `%s_raw`.
        clip : bool, optional
            If True, negative signal will be clipped to 0. (default to False )
        vars : list, optional
            variables names to add, by default `['sigma0' , 'beta0' , 'gamma0']`

        Returns
        -------
        xarray.DataSet
            dataset with denoised vars
        """
        if vars is None:
            vars = ['sigma0', 'beta0', 'gamma0']
        for varname in vars:
            varname_raw = varname + '_raw'
            noise = 'ne%sz' % varname[0]
            if varname_raw not in ds:
                continue
            else:
                denoised = ds[varname_raw] - ds[noise]

                if clip:
                    denoised = denoised.clip(min=0)
                    denoised.attrs['comment'] = 'clipped, no values <0'
                else:
                    denoised.attrs['comment'] = 'not clipped, some values can be <0'
                ds[varname] = denoised
        return ds

    @timing
    def _apply_calibration_lut(self, var_name):
        """
            Apply calibration lut to `digital_number` to compute `var_name`.

            Parameters
            ----------
            var_name: str
                Variable name to compute by applying lut. Must exist in `self._map_var_lut` to be able to get the corresponding lut.

            Returns
            -------
            xarray.Dataset
                with one variable named by `var_name`
        """
        lut = self._get_lut(var_name)
        offset = lut.attrs['offset']
        # if self.resolution is not None:
        lut = self._resample_lut_values(lut)
        res = ((self._dataset.digital_number ** 2.) + offset) / lut
        # Have to know if we keep this line written by Olivier because it replaces 0 values by nan --> creates problems for wind inversion
        res = res.where(res > 0)
        res.attrs.update(lut.attrs)
        return res.to_dataset(name=var_name + '_raw')

    @timing
    def flip_sample_da(self, ds):
        """
        When a product is flipped, flip back data arrays (from a dataset) sample dimensions to respect the xsar
        convention (increasing incidence values)

        Parameters
        ----------
        ds : xarray.Dataset
            Contains dataArrays which depends on `sample` dimension

        Returns
        -------
        xarray.Dataset
            Flipped back, respecting the xsar convention
        """
        antenna_pointing = self.sar_meta.dt['radarParameters'].attrs['antennaPointing']
        pass_direction = self.sar_meta.dt.attrs['passDirection']
        flipped_cases = [('Left', 'Ascending'), ('Right', 'Descending')]
        if (antenna_pointing, pass_direction) in flipped_cases:
            new_ds = ds.copy().isel(sample=slice(None, None, -1)
                                    ).assign_coords(sample=ds.sample)
        else:
            new_ds = ds
        return new_ds

    @timing
    def flip_line_da(self, ds):
        """
        Flip dataArrays (from a dataset) that depend on line dimension when a product is ascending, in order to
        respect the xsar convention (increasing time along line axis, whatever ascending or descending product).
        Reference : `schemas/rs2prod_burstAttributes.xsd:This corresponds to the top-left pixel in a coordinate
        system where the range increases to the right and the zero-Doppler time increases downward. Note that this is
        not necessarily the top-left pixel of the image block in the final product.`

        Parameters
        ----------
        ds : xarray.Dataset
            Contains dataArrays which depends on `line` dimension

        Returns
        -------
        xarray.Dataset
            Flipped back, respecting the xsar convention
        """
        pass_direction = self.sar_meta.dt.attrs['passDirection']
        if pass_direction == 'Ascending':
            new_ds = ds.copy().isel(line=slice(None, None, -1)).assign_coords(line=ds.line)
        else:
            new_ds = ds.copy()
        return new_ds

    @property
    def interpolate_times(self):
        """
        Apply interpolation with RectBivariateSpline to the azimuth time extracted from `self.sar_meta.geoloc`

        Returns
        -------
        xarray.Dataset
            Contains the time as delayed at the good resolution and expressed as type datetime64[ns]
        """
        times = self.sar_meta.get_azitime
        lines = self.sar_meta.geoloc.line
        samples = self.sar_meta.geoloc.pixel
        time_values_2d = np.tile(times, (samples.shape[0], 1)).transpose()
        interp_func = RectBivariateSpline(
            x=lines, y=samples, z=time_values_2d.astype(float), kx=1, ky=1)
        da_var = map_blocks_coords(
            self._da_tmpl.astype('datetime64[ns]'),
            interp_func
        )
        return da_var.isel(sample=0).to_dataset(name='time')

    def get_sensor_velocity(self):
        """
        Interpolated sensor velocity
        Returns
        -------
        xarray.Dataset()
            containing a single variable velocity
        """
        azimuth_times = self.sar_meta.get_azitime
        interp_times = self.interpolate_times['time']
        orbstatevect = self.sar_meta.orbit_and_attitude
        orbstatevect2 = xr.Dataset()
        # number of values must be the same as number of lines
        for var in orbstatevect:
            da = orbstatevect[var]
            interp_func = interp1d(y=da, x=da.timeStamp.astype('float'))
            interp_var = interp_func(azimuth_times.astype('float'))
            orbstatevect2[var] = xr.DataArray(data=interp_var, dims=['timeStamp'], coords={'timeStamp': azimuth_times},
                                              attrs=da.attrs)
        orbstatevect2.attrs = orbstatevect.attrs
        velos = np.array(
            [orbstatevect2['xVelocity'] ** 2., orbstatevect2['yVelocity'] ** 2., orbstatevect2['zVelocity'] ** 2.])
        vels = np.sqrt(np.sum(velos, axis=0))
        interp_f = interp1d(azimuth_times.astype(float), vels)
        _vels = interp_f(interp_times.astype(float))
        res = xr.DataArray(_vels, dims=['line'], coords={
                           'line': self.dataset.line})
        return xr.Dataset({'velocity': res})

    def _reconfigure_reader_datatree(self):
        """
        Modify the structure of self.datatree to merge the reader's one and the measurement dataset.
        Modify the structure of the reader datatree for a better user experience.
        Include attributes from the reader (concerning the satellite) in the attributes of self.datatree

        Returns
        -------
        datatree.Datatree
        """

        dic = {'measurement': self.datatree['measurement'],
               'geolocation_annotation': self.datatree['geolocation_annotation'],
               }

        def del_items_in_dt(dt, list_items):
            for item in list_items:
                dt.to_dataset().__delitem__(item)
            return

        def get_list_keys_delete(dt, list_keys, inside=True):
            dt_keys = dt.keys()
            final_list = []
            for item in dt_keys:
                if inside:
                    if item in list_keys:
                        final_list.append(item)
                else:
                    if item not in list_keys:
                        final_list.append(item)
            return final_list

        exclude = ['geolocationGrid', 'lut', 'radarParameters']
        rename_lut = {
            'lutBeta': 'beta0_lut',
            'lutGamma': 'gamma0_lut',
            'lutSigma': 'sigma0_lut',
        }
        rename_radarParameters = {
            'noiseLevelValues_BetaNought': 'beta0_noise_lut',
            'noiseLevelValues_SigmaNought': 'sigma0_noise_lut',
            'noiseLevelValues_Gamma': 'gamma0_noise_lut'
        }
        new_dt = datatree.DataTree.from_dict(dic)

        dt = self.sar_meta.dt

        # rename lut
        new_dt['lut'] = dt['lut'].rename(rename_lut)

        # extract noise_lut, rename and put these in a dataset
        new_dt['noise_lut'] = dt['radarParameters'].rename(
            rename_radarParameters)
        new_dt['noise_lut'].attrs = {}  # reset attributes
        delete_list = get_list_keys_delete(
            new_dt['noise_lut'], rename_radarParameters.values(), inside=False)
        del_items_in_dt(new_dt['noise_lut'], delete_list)

        # Create a dataset for radar parameters without the noise luts
        new_dt['radarParameters'] = dt['radarParameters']
        delete_list = get_list_keys_delete(
            new_dt['radarParameters'], rename_radarParameters.keys())
        del_items_in_dt(new_dt['radarParameters'], delete_list)

        # extract others dataset
        copy_dt = dt.copy()
        for key in dt.copy():
            if key not in exclude:
                if key == 'imageGenerationParameters':
                    new_dt[key] = datatree.DataTree(
                        parent=None, children=copy_dt[key])
                else:
                    new_dt[key] = copy_dt[key]
        self.datatree = new_dt
        self.datatree.attrs.update(self.sar_meta.dt.attrs)
        return

    def __repr__(self):
        if self.sliced:
            intro = "sliced"
        else:
            intro = "full coverage"
        return "<RadarSat2Dataset %s object>" % intro

    @property
    def dataset(self):
        """
        `xarray.Dataset` representation of this `xsar.RadarSat2Dataset` object.
        This property can be set with a new dataset, if the dataset was computed from the original dataset.
        """
        # return self._dataset
        res = self.datatree['measurement'].to_dataset()
        res.attrs = self.datatree.attrs
        return res

    @dataset.setter
    def dataset(self, ds):
        if self.sar_meta.name == ds.attrs['name']:
            # check if new ds has changed coordinates
            if not self.sliced:
                self.sliced = any(
                    [list(ds[d].values) != list(self._dataset[d].values) for d in ['line', 'sample']])
            self._dataset = ds
            # self._dataset = self.datatree['measurement'].ds
            self.recompute_attrs()
        else:
            raise ValueError("dataset must be same kind as original one.")

    @dataset.deleter
    def dataset(self):
        logger.debug('deleter dataset')

    @property
    def rs2meta(self):
        logger.warning('Please use `sar_meta` to call the sar meta object')
        return self.sar_meta



import copy
import logging
import warnings

import cartopy
import rasterio
import shapely
from shapely.geometry import Polygon
from shapely import ops
import numpy as np
from datetime import datetime

from abc import abstractmethod

from .raster_readers import available_rasters
from .base_dataset import BaseDataset
import geopandas as gpd

from .utils import class_or_instancemethod, to_lon180, haversine

logger = logging.getLogger('xsar.base_meta')
logger.addHandler(logging.NullHandler())

# we know tiff as no geotransform : ignore warning
warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)

# allow nan without warnings
# some dask warnings are still non filtered: https://github.com/dask/dask/issues/3245
np.errstate(invalid='ignore')


class BaseMeta(BaseDataset):
    """
        Abstract class that defines necessary common functions for the computation of different SAR metadata
        (Radarsat2, Sentinel1, RCM...).
        This also permit a better maintenance, because these functions aren't redefined many times.
    """

    # default mask feature (see self.set_mask_feature and cls.set_mask_feature)
    _mask_features_raw = {
        'land': cartopy.feature.NaturalEarthFeature('physical', 'land', '10m')
    }

    rasters = available_rasters.iloc[0:0].copy()

    _mask_features = {}
    _mask_intersecting_geometries = {}
    _mask_geometry = {}
    _geoloc = None
    _rasterized_masks = None
    manifest_attrs = None
    _time_range = None
    name = None
    multidataset = None
    short_name = None
    path = None
    product = None
    manifest = None
    subdatasets = None
    dsid = None
    safe = None
    geoloc = None

    def _get_mask_feature(self, name):
        # internal method that returns a cartopy feature from a mask name
        if self._mask_features[name] is None:
            feature = self._mask_features_raw[name]
            if isinstance(feature, str):
                # feature is a shapefile.
                # we get the crs from the shapefile to be able to transform the footprint to this crs_in
                # (so we can use `mask=` in gpd.read_file)
                import fiona
                import pyproj
                from shapely.ops import transform
                with fiona.open(feature) as fshp:
                    try:
                        # proj6 give a " FutureWarning: '+init=<authority>:<code>' syntax is deprecated. "
                        # '<authority>:<code>' is the preferred initialization method"
                        crs_in = fshp.crs['init']
                    except KeyError:
                        crs_in = fshp.crs
                    crs_in = pyproj.CRS(crs_in)
                proj_transform = pyproj.Transformer.from_crs(pyproj.CRS('EPSG:4326'), crs_in, always_xy=True).transform
                footprint_crs = transform(proj_transform, self.footprint)

                with warnings.catch_warnings():
                    # ignore "RuntimeWarning: Sequential read of iterator was interrupted. Resetting iterator."
                    warnings.simplefilter("ignore", RuntimeWarning)
                    feature = cartopy.feature.ShapelyFeature(
                        gpd.read_file(feature, mask=footprint_crs).to_crs(epsg=4326).geometry,
                        cartopy.crs.PlateCarree()
                    )
            if not isinstance(feature, cartopy.feature.Feature):
                raise TypeError('Expected a cartopy.feature.Feature type')
            self._mask_features[name] = feature

        return self._mask_features[name]

    @class_or_instancemethod
    def set_mask_feature(self_or_cls, name, feature):
        """
        Set a named mask from a shapefile or a cartopy feature.

        Parameters
        ----------
        name: str
            mask name
        feature: str or cartopy.feature.Feature
            if str, feature is a path to a shapefile or whatever file readable with fiona.
            It is recommended to use str, as the serialization of cartopy feature might be big.

        Examples
        --------
            Add an 'ocean' mask at class level (ie as default mask):
            ```
            >>> xsar.RadarSat2Meta.set_mask_feature('ocean', cartopy.feature.OCEAN)
            >>> xsar.Sentinel1Meta.set_mask_feature('ocean', cartopy.feature.OCEAN)
            ```

            Add an 'ocean' mask at instance level (ie only for this self Sentinel1Meta (or RadarSat2Meta instance):
            ```
            >>> xsar.RadarSat2Meta.set_mask_feature('ocean', cartopy.feature.OCEAN)
            >>> xsar.Sentinel1Meta.set_mask_feature('ocean', cartopy.feature.OCEAN)
            ```


            High resoltion shapefiles can be found from openstreetmap.
            It is recommended to use WGS84 with large polygons split from https://osmdata.openstreetmap.de/

        See Also
        --------
        xsar.BaseMeta.get_mask
        """

        # see https://stackoverflow.com/a/28238047/5988771 for self_or_cls

        self_or_cls._mask_features_raw[name] = feature

        if not isinstance(self_or_cls, type):
            # self (instance, not class)
            self_or_cls._mask_intersecting_geometries[name] = None
            self_or_cls._mask_geometry[name] = None
            self_or_cls._mask_features[name] = None

    def get_mask(self, name, describe=False):
        """
        Get mask from `name` (e.g. 'land') as a shapely Polygon.
        The resulting polygon is contained in the footprint.

        Parameters
        ----------
        name: str

        Returns
        -------
        shapely.geometry.Polygon

        """
        if describe:
            descr = self._mask_features_raw[name]
            try:
                # nice repr for a class (like 'cartopy.feature.NaturalEarthFeature land')
                descr = '%s.%s %s' % (descr.__module__, descr.__class__.__name__, descr.name)
            except AttributeError:
                pass
            return descr

        if self._mask_geometry[name] is None:
            if self._get_mask_intersecting_geometries(name).unary_union:
                poly = self._get_mask_intersecting_geometries(name) \
                    .unary_union.intersection(self.footprint)

            else:
                poly = Polygon()

            if poly.is_empty:
                poly = Polygon()

            self._mask_geometry[name] = poly
        return self._mask_geometry[name]

    def _get_mask_intersecting_geometries(self, name):
        """

        :param name: str(eg land)
        :return:
        """
        if self._mask_intersecting_geometries[name] is None:
            gseries = gpd.GeoSeries(self._get_mask_feature(name).geometries())
            # gseries = gpd.GeoSeries(self._get_mask_feature(name)
            #                         .intersecting_geometries(self.footprint))
            if len(gseries) == 0:
                # no intersection with mask, but we want at least one geometry in the serie (an empty one)
                gseries = gpd.GeoSeries([Polygon()])
            self._mask_intersecting_geometries[name] = gseries
        return self._mask_intersecting_geometries[name]

    @property
    @abstractmethod
    def footprint(self):
        pass

    @property
    def cross_antemeridian(self):
        """True if footprint cross antemeridian"""
        return ((np.max(self.geoloc['longitude']) - np.min(
            self.geoloc['longitude'])) > 180).item()

    @property
    def swath(self):
        """string like 'EW', 'IW', 'WV', etc ..."""
        return self.manifest_attrs['swath_type']

    @property
    @abstractmethod
    def _dict_coords2ll(self):
        pass

    @property
    @abstractmethod
    def approx_transform(self):
        pass

    @property
    def mask_names(self):
        """

        Returns
        -------
        list of str
            mask names
        """
        return self._mask_features.keys()

    def coords2ll(self, *args, to_grid=False, approx=False):
        """
        convert `lines`, `samples` arrays to `longitude` and `latitude` arrays.
        or a shapely object in `lines`, `samples` coordinates to `longitude` and `latitude`.

        Parameters
        ----------
        *args: lines, samples  or a shapely geometry
            lines, samples are iterables or scalar

        to_grid: bool, default False
            If True, `lines` and `samples` must be 1D arrays. The results will be 2D array of shape (lines.size, samples.size).

        Returns
        -------
        tuple of np.array or tuple of float
            (longitude, latitude) , with shape depending on `to_grid` keyword.

        See Also
        --------
        xsar.BaseMeta.ll2coords
        xsar.BaseDataset.ll2coords

        """
        if isinstance(args[0], shapely.geometry.base.BaseGeometry):
            return self._coords2ll_shapely(args[0])

        lines, samples = args

        scalar = True
        if hasattr(lines, '__iter__'):
            scalar = False

        if approx:
            if to_grid:
                samples2D, lines2D = np.meshgrid(samples, lines)
                lon, lat = self.approx_transform * (lines2D, samples2D)
                pass
            else:
                lon, lat = self.approx_transform * (lines, samples)
        else:
            dict_coords2ll = self._dict_coords2ll
            if to_grid:
                lon = dict_coords2ll['longitude'](lines, samples)
                lat = dict_coords2ll['latitude'](lines, samples)
            else:
                lon = dict_coords2ll['longitude'].ev(lines, samples)
                lat = dict_coords2ll['latitude'].ev(lines, samples)

        if self.cross_antemeridian:
            lon = to_lon180(lon)

        if scalar and hasattr(lon, '__iter__'):
            lon = lon.item()
            lat = lat.item()

        if hasattr(lon, '__iter__') and type(lon) is not type(lines):
            lon = type(lines)(lon)
            lat = type(lines)(lat)

        return lon, lat

    def _ll2coords_shapely(self, shape, approx=False):
        if approx:
            (xoff, a, b, yoff, d, e) = (~self.approx_transform).to_gdal()
            return shapely.affinity.affine_transform(shape, (a, b, d, e, xoff, yoff))
        else:
            return shapely.ops.transform(self.ll2coords, shape)

    def _coords2ll_shapely(self, shape, approx=False):
        if approx:
            (xoff, a, b, yoff, d, e) = self.approx_transform.to_gdal()
            return shapely.affinity.affine_transform(shape, (a, b, d, e, xoff, yoff))
        else:
            return shapely.ops.transform(self.coords2ll, shape)

    def ll2coords(self, *args):
        """
        Get `(lines, samples)` from `(lon, lat)`,
        or convert a lon/lat shapely object to line/sample coordinates.

        Parameters
        ----------
        *args: lon, lat or shapely object
            lon and lat might be iterables or scalars

        Returns
        -------
        tuple of np.array or tuple of float (lines, samples) , or a shapely object

        Examples
        --------
            get nearest (line,sample) from (lon,lat) = (84.81, 21.32) in ds, without bounds checks

            >>> (line, sample) = self.ll2coords(84.81, 21.32) # (lon, lat)
            >>> (line, sample)
            (9752.766349989339, 17852.571322887554)

        See Also
        --------
        xsar.BaseMeta.coords2ll
        xsar.BaseDataset.coords2ll

        """

        if isinstance(args[0], shapely.geometry.base.BaseGeometry):
            return self._ll2coords_shapely(args[0])

        lon, lat = args

        # approximation with global inaccurate transform
        line_approx, sample_approx = ~self.approx_transform * (np.asarray(lon), np.asarray(lat))

        # Theoretical identity. It should be the same, but the difference show the error.
        lon_identity, lat_identity = self.coords2ll(line_approx, sample_approx, to_grid=False)
        line_identity, sample_identity = ~self.approx_transform * (lon_identity, lat_identity)

        # we are now able to compute the error, and make a correction
        line_error = line_identity - line_approx
        sample_error = sample_identity - sample_approx

        line = line_approx - line_error
        sample = sample_approx - sample_error

        if hasattr(lon, '__iter__'):
            scalar = False
        else:
            scalar = True

        return line, sample

    def coords2heading(self, lines, samples, to_grid=False, approx=True):
        """
        Get image heading (lines increasing direction) at coords `lines`, `samples`.

        Parameters
        ----------
        lines: np.array or scalar
        samples: np.array or scalar
        to_grid: bool
            If True, `lines` and `samples` must be 1D arrays. The results will be 2D array of shape (lines.size, samples.size).

        Returns
        -------
        np.array or float
            `heading` , with shape depending on `to_grid` keyword.

        """

        lon1, lat1 = self.coords2ll(lines - 1, samples, to_grid=to_grid, approx=approx)
        lon2, lat2 = self.coords2ll(lines + 1, samples, to_grid=to_grid, approx=approx)
        _, heading = haversine(lon1, lat1, lon2, lat2)
        return heading

    @property
    @abstractmethod
    def _get_time_range(self):
        pass

    @property
    def time_range(self):
        """time range as pd.Interval"""
        if self._time_range is None:
            self._time_range = self._get_time_range()
        return self._time_range

    @property
    def start_date(self):
        """start date, as datetime.datetime"""
        out_format = '%Y-%m-%d %H:%M:%S.%f'
        date = self.time_range.left
        try:
            return '%s' % datetime.strptime('%s' % date, out_format)
        except ValueError:
            return '%s' % date.strftime(out_format)

    @property
    def stop_date(self):
        """stop date, as datetime.datetime"""
        out_format = '%Y-%m-%d %H:%M:%S.%f'
        date = self.time_range.right
        try:
            return '%s' % datetime.strptime('%s' % date, out_format)
        except ValueError:
            return '%s' % date.strftime(out_format)

    @class_or_instancemethod
    def set_raster(self_or_cls, name, resource, read_function=None, get_function=None):
        # get defaults if exists
        default = available_rasters.loc[name:name]

        # set from params, or from default
        self_or_cls.rasters.loc[name, 'resource'] = resource or default.loc[name, 'resource']
        self_or_cls.rasters.loc[name, 'read_function'] = read_function or default.loc[name, 'read_function']
        self_or_cls.rasters.loc[name, 'get_function'] = get_function or default.loc[name, 'get_function']

        return

    @property
    def dict(self):
        # return a minimal dictionary that can be used with Sentinel1Meta.from_dict() or pickle (see __reduce__)
        # to reconstruct another instance of self
        #
        minidict = {
            'name': self.name,
            '_mask_features_raw': self._mask_features_raw,
            '_mask_features': {},
            '_mask_intersecting_geometries': {},
            '_mask_geometry': {},
            'rasters': self.rasters
        }
        for name in minidict['_mask_features_raw'].keys():
            minidict['_mask_intersecting_geometries'][name] = None
            minidict['_mask_geometry'][name] = None
            minidict['_mask_features'][name] = None
        return minidict

    @classmethod
    def from_dict(cls, minidict):
        # like copy constructor, but take a dict from Sentinel1Meta.dict
        # https://github.com/umr-lops/xsar/issues/23
        for name in minidict['_mask_features_raw'].keys():
            assert minidict['_mask_geometry'][name] is None
            assert minidict['_mask_features'][name] is None
        minidict = copy.copy(minidict)
        new = cls(minidict['name'])
        new.__dict__.update(minidict)
        return new


import pandas as pd
import rasterio
from rasterio.control import GroundControlPoint
from scipy.interpolate import RectBivariateSpline, interp1d
from shapely.geometry import Polygon

from .utils import haversine, timing
import os
import geopandas as gpd
import numpy as np
from .base_meta import BaseMeta


class RadarSat2Meta(BaseMeta):
    """
        Handle dataset metadata.
        A `xsar.RadarSat2Meta` object can be used with `xsar.open_dataset`,
        but it can be used as itself: it contains usefull attributes and methods.

        Parameters
        ----------
        name: str
            path or gdal identifier

    """
    # class attributes are needed to fetch instance attribute (ie self.name) with dask actors
    # ref http://distributed.dask.org/en/stable/actors.html#access-attributes
    # FIXME: not needed if @property, so it might be a good thing to have getter for those attributes
    dt = None

    @timing
    def __init__(self, name):
        from xradarsat2 import rs2_reader
        if ':' in name:
            self.dt = rs2_reader(name.split(':')[1])
        else:
            self.dt = rs2_reader(name)
        if not name.startswith('RADARSAT2_DS:'):
            name = 'RADARSAT2_DS:%s:' % name
        self.name = name
        """Gdal dataset name"""
        name_parts = self.name.split(':')
        if len(name_parts) > 3:
            # windows might have semicolon in path ('c:\...')
            name_parts[1] = ':'.join(name_parts[1:-1])
            del name_parts[2:-1]
        name_parts[1] = os.path.basename(name_parts[1])
        self.short_name = ':'.join(name_parts)
        """Like name, but without path"""
        self.path = ':'.join(self.name.split(':')[1:-1])
        """Dataset path"""
        self.safe = os.path.basename(self.path)
        """Safe file name"""
        # there is no information on resolution 'F' 'H' or 'M' in the manifest, so we have to extract it from filename
        try:
            self.product = os.path.basename(self.path).split('_')[9]
        except:
            self.product = "XXX"
        """Product type, like 'GRDH', 'SLC', etc .."""

        # Respect xsar convention flipping lines and samples from the reader when necessary
        self.samples_flipped = False
        self.lines_flipped = False
        self.flip_line_da()
        self.flip_sample_da()

        self._safe_files = None
        self.multidataset = False
        """True if multi dataset"""
        self.subdatasets = gpd.GeoDataFrame(geometry=[], index=[])
        """Subdatasets as GeodataFrame (empty if single dataset)"""
        self.geoloc = self.dt['geolocationGrid'].to_dataset()

        self.orbit_and_attitude = self.dt['orbitAndAttitude'].ds
        self.doppler_centroid = self.dt['imageGenerationParameters']['doppler']['dopplerCentroid'].ds
        self.doppler_rate_values = self.dt['imageGenerationParameters']['doppler']['dopplerRateValues'].ds
        self.chirp = self.dt['imageGenerationParameters']['chirp'].ds
        self.radar_parameters = self.dt['radarParameters'].ds
        self.lut = self.dt['lut'].ds
        self.manifest_attrs = self._create_manifest_attrs()
        for name, feature in self.__class__._mask_features_raw.items():
            self.set_mask_feature(name, feature)

    def _create_manifest_attrs(self):
        dic = dict()
        dic["swath_type"] = os.path.basename(self.path).split('_')[4]
        dic["polarizations"] = self.dt["radarParameters"]["pole"].values
        dic["product_type"] = self.product
        dic['satellite'] = self.dt.attrs['satellite']
        dic['start_date'] = self.start_date
        dic['stop_date'] = self.stop_date
        # compute attributes (footprint, coverage, pixel_size)
        footprint_dict = {}
        for ll in ['longitude', 'latitude']:
            footprint_dict[ll] = [
                self.geoloc[ll].isel(line=a, pixel=x).values for a, x in [(0, 0), (0, -1), (-1, -1), (-1, 0)]
            ]
        corners = list(zip(footprint_dict['longitude'], footprint_dict['latitude']))
        p = Polygon(corners)
        self.geoloc.attrs['footprint'] = p
        dic["footprints"] = p
        # compute acquisition size/resolution in meters
        # first vector is on sample
        acq_sample_meters, _ = haversine(*corners[0], *corners[1])
        # second vector is on line
        acq_line_meters, _ = haversine(*corners[1], *corners[2])
        dic['coverage'] = "%dkm * %dkm (line * sample )" % (
            acq_line_meters / 1000, acq_sample_meters / 1000)

        def _to_rio_gcp(pt_geoloc):
            # convert a point from self._geoloc grid to rasterio GroundControlPoint
            return GroundControlPoint(
                x=pt_geoloc.longitude.item(),
                y=pt_geoloc.latitude.item(),
                z=pt_geoloc.height.item(),
                col=pt_geoloc.line.item(),
                row=pt_geoloc.pixel.item()
            )

        gcps = [
            _to_rio_gcp(self.geoloc.sel(line=line, pixel=sample))
            for line in self.geoloc.line for sample in self.geoloc.pixel
        ]
        # approx transform, from all gcps (inaccurate)
        dic['approx_transform'] = rasterio.transform.from_gcps(gcps)
        return dic

    @property
    def approx_transform(self):
        """
        Affine transfom from geoloc.

        This is an inaccurate transform, with errors up to 600 meters.
        But it's fast, and may fit some needs, because the error is stable localy.
        See `xsar.BaseMeta.coords2ll` `xsar.BaseMeta.ll2coords` for accurate methods.

        Examples
        --------
            get `longitude` and `latitude` from tuple `(line, sample)`:

            >>> longitude, latitude = self.approx_transform * (line, sample)

            get `line` and `sample` from tuple `(longitude, latitude)`

            >>> line, sample = ~self.approx_transform * (longitude, latitude)

        See Also
        --------
        xsar.BaseDataset.coords2ll
        xsar.BaseDataset.ll2coords
        xsar.BaseMeta.coords2ll
        xsar.BaseMeta.ll2coords

        """
        return self.manifest_attrs['approx_transform']

    def to_dict(self, keys='minimal'):

        info_keys = {
            'minimal': [
                #'platform',
                'swath', 'product', 'pols']
        }
        info_keys['all'] = info_keys['minimal'] + ['name', 'start_date', 'stop_date',
                                                   'footprint',
                                                   'coverage',
                                                   'pixel_line_m', 'pixel_sample_m',
                                                   'approx_transform',

                                                   #'orbit_pass',
                                                   #'platform_heading'
                                                   ]

        if isinstance(keys, str):
            keys = info_keys[keys]

        res_dict = {}
        for k in keys:
            if hasattr(self, k):
                res_dict[k] = getattr(self, k)
            elif k in self.manifest_attrs.keys():
                res_dict[k] = self.manifest_attrs[k]
            else:
                raise KeyError('Unable to find key/attr "%s" in RadarSat2Meta' % k)
        return res_dict

    @property
    def pols(self):
        """polarisations strings, separated by spaces """
        return " ".join(self.manifest_attrs['polarizations'])

    @property
    def footprint(self):
        """footprint, as a shapely polygon or multi polygon"""
        return self.geoloc.attrs['footprint']

    @property
    def pixel_line_m(self):
        """pixel line spacing, in meters (at sensor level)"""
        if self.multidataset:
            res = None  # not defined for multidataset
        else:
            res = self.geoloc.line.attrs['rasterAttributes_sampledLineSpacing_value']
        return res

    @property
    def pixel_sample_m(self):
        """pixel sample spacing, in meters (at sensor level)"""
        if self.multidataset:
            res = None  # not defined for multidataset
        else:
            res = self.geoloc.pixel.attrs['rasterAttributes_sampledPixelSpacing_value']
        return res

    def _get_time_range(self):
        if self.multidataset:
            time_range = [self.manifest_attrs['start_date'], self.manifest_attrs['stop_date']]
        else:
            time_range = self.orbit_and_attitude.timeStamp
        return pd.Interval(left=pd.Timestamp(time_range.values[0]), right=pd.Timestamp(time_range.values[-1]), closed='both')

    @property
    def get_azitime(self):
        """
        Get time at low resolution

        Returns
        -------
        array[datetime64[ns]]
            times
        """
        times = self.orbit_and_attitude.timeStamp.values
        lines = self.geoloc.line
        # Some products have a different number of lines and timestamps
        if len(times) != len(lines):
            interp_func = interp1d(y=times, x=np.arange(1, len(times) + 1))
            times = interp_func(np.linspace(1, len(times), len(lines)))
        return times

    def __reduce__(self):
        # make self serializable with pickle
        # https://docs.python.org/3/library/pickle.html#object.__reduce__

        return self.__class__, (self.name,), self.dict

    def __repr__(self):
        if self.multidataset:
            meta_type = "multi (%d)" % len(self.subdatasets)
        else:
            meta_type = "single"
        return "<RadarSat2Meta %s object>" % meta_type

    @property
    def _dict_coords2ll(self):
        """
        dict with keys ['longitude', 'latitude'] with interpolation function (RectBivariateSpline) as values.

        Examples:
        ---------
            get longitude at line=100 and sample=200:
            ```
            >>> self._dict_coords2ll['longitude'].ev(100,200)
            array(-66.43947434)
            ```
        Notes:
        ------
            if self.cross_antemeridian is True, 'longitude' will be in range [0, 360]
        """
        resdict = {}
        geoloc = self.geoloc
        if self.cross_antemeridian:
            geoloc['longitude'] = geoloc['longitude'] % 360

        idx_sample = np.array(geoloc.pixel)
        idx_line = np.array(geoloc.line)

        for ll in ['longitude', 'latitude']:
            resdict[ll] = RectBivariateSpline(idx_line, idx_sample, np.asarray(geoloc[ll]), kx=1, ky=1)

        return resdict

    def flip_sample_da(self):
        """
        When a product is flipped, flip back data arrays (from the reader datatree) sample dimensions to respect the xsar
        convention (increasing incidence values)
        """
        antenna_pointing = self.dt['radarParameters'].attrs['antennaPointing']
        pass_direction = self.dt.attrs['passDirection']
        flipped_cases = [('Left', 'Ascending'), ('Right', 'Descending')]
        samples_depending_ds = ['geolocationGrid', 'lut', 'radarParameters']
        if (antenna_pointing, pass_direction) in flipped_cases:
            for ds_name in samples_depending_ds:
                if 'radar' in ds_name:
                    self.dt[ds_name] = self.dt[ds_name].rename({'NbOfNoiseLevelValues': 'pixel'})
                self.dt[ds_name] = self.dt[ds_name].copy().isel(pixel=slice(None, None, -1))\
                    .assign_coords(pixel=self.dt[ds_name].ds.pixel)
            self.samples_flipped = True
        return

    def flip_line_da(self):
        """
        Flip dataArrays (from the reader datatree) that depend on line dimension when a product is ascending, in order to
        respect the xsar convention (increasing time along line axis, whatever ascending or descending product).
        Reference : `schemas/rs2prod_burstAttributes.xsd:This corresponds to the top-left pixel in a coordinate
        system where the range increases to the right and the zero-Doppler time increases downward. Note that this is
        not necessarily the top-left pixel of the image block in the final product.`
        """
        pass_direction = self.dt.attrs['passDirection']
        samples_depending_ds = ['geolocationGrid']
        if pass_direction == 'Ascending':
            for ds_name in samples_depending_ds:
                self.dt[ds_name] = self.dt[ds_name].copy().isel(line=slice(None, None, -1))\
                    .assign_coords(line=self.dt[ds_name].ds.line)
            self.lines_flipped = True
        return



